
;  Lcd driver for the PIC 18F series

;  Four examples of how data can be output

;  Configured here for the 4 mhz internal oscillator.
;  
;  (different frequencies will need the software delays modifying )


		LIST P=18F4550, F=INHX32		;directive to define processor
		#include <P18F4550.INC>			;processor specific variable definitions



	
		CONFIG WDT=OFF					; disable watchdog timer
		CONFIG MCLRE = ON				; MCLEAR Pin on
		CONFIG DEBUG = OFF				; Enable Debug Mode
		CONFIG LVP = OFF				; Disable low voltage programming
		CONFIG OSC=INTIO67				; internal oscillator  4520/2520
;		CONFIG FOSC = INTOSCIO_EC		; internal oscillator  4550/2550
;		CONFIG OSC = IRCIO67 			; internal oscillator  4580
		CONFIG PBADEN = OFF             ; PORTB all digital



; Select your lcd port here..

#define LCDPORT PORTB
#define LCDTRIS TRISB

;#define LCDPORT PORTD
;#define LCDTRIS TRISD



        ;Variable definitions
       
        CBLOCK	0x000
		lcd_flags,lcd_RSflg,lcd_temp
		d1,d2,d3
     	ENDC

       	ORG	0x0000						;Reset vector This code will start executing when a reset occurs.
		goto Main

;******************************************************************************
;High priority interrupt vector
		ORG	0x0008
		retfie

;******************************************************************************
;Low priority interrupt vector and routine
;
		ORG	0x0018
		retfie


Main	
		movlw	0x07
		movwf	CMCON	 			;turn comparators off (make it like a 16F84)
		movlw 	0x0F		 		; turn off all analog pins
		movwf 	ADCON1
   	
		clrf 	PORTB			
   		clrf	TRISB				; PortB as D outputs

		movlw	b'01100010'			; set internal osc to 4 mhz
		movwf	OSCCON


		call 	lcd_init			; Initialise the LCD module



;******************************************************************************

; Example 1   - sending out two rows of data

; Ok for testing or very simple data dispays,  but not very practical
; see example 4 for a better solution.

direct_text							; send out two lines of data to lcd
	    movlw	0x80				; add address for top line
		call    lcd_c				; send out command
			
		movlw	'h'					; now send out data
		call	lcd_d				
		movlw	'e'
		call	lcd_d
		movlw	'l'
		call	lcd_d
		movlw	'l'
		call	lcd_d
		movlw	'o'
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d		
		
												 		
		movlw	0xC0				; add address for bottom line	
		call    lcd_c	
		
		movlw	'W'					; now send out data
		call	lcd_d
		movlw	'o'
		call	lcd_d
		movlw	'r'
		call	lcd_d
		movlw	'l'
		call	lcd_d
		movlw	'd'
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		movlw	' '
		call	lcd_d
		


		call	delay1sec			; finished display, little pause
		call	delay1sec
		call	delay1sec
		call	delay1sec
		call	delay1sec


;******************************************************************************

; 	Example 2    Sending out two rows of data held in 'Strings' 

; 	here two strings are placed directly in the code, 


string_example

	    movlw	0x80				; add address for top line
		call    lcd_c
        call    PutString       
        dw      "String Top",0      ; string must be terminated with a 0 byte
	    movlw	0xC0				; add address for bottom line
		call    lcd_c
        call    PutString       
        dw      "String Bottom...",0   ; string must be terminated with a 0 byte

		call	delay1sec			; finished display, little pause
		call	delay1sec
		call	delay1sec
		call	delay1sec
		call	delay1sec

  
		goto  	string_example3


; the PutString sub-routine by Mike McClaren, K8LH
; http://www.electro-tech-online.com/microcontrollers/94132-new-18f-memory-look-up-table-questions.html
	
;  PutString - print in-line string via Stack and TBLPTR
;
;  string must be terminated with a 0 byte and does not need
;  to be word aligned



PutString
        movff   TOSL,TBLPTRL    	; copy return address into TBLPTR
        movff   TOSH,TBLPTRH    
        clrf    TBLPTRU         	; assume PIC with < 64-KB
PutNext
        tblrd   *+              	; get in-line string character
        movf    TABLAT,W        	; last character (0)?
        bz      PutExit         	; yes, exit, else
        call   	lcd_d        		; print character
        bra     PutNext         	; and do another
PutExit
        btfsc   TBLPTRL,0       	; odd address?
        tblrd   *+              	; yes, make it even (fix PC)
        movf    TBLPTRH,W       	; setup new return address
        movwf   TOSH            	
        movf    TBLPTRL,W       
        movwf   TOSL            
        return                  	; to 1st instruction after the string table



;******************************************************************************

; 	Example 3    Sending out a Line of String data from anywhere in your code using a Macro
;   Macros PutStr1 and 2 placed at the top of your main code, must come before they are calld


string_example3  nop

PutStr1	macro   string1
	    movlw	0x80				; add address for Top line
		call    lcd_c
        call    PutString       
        dw      string1,0        	; null terminated in-line string table
        endm   

PutStr2	macro   string2
	    movlw	0xC0				; add address for top line
		call    lcd_c
        call    PutString       
        dw      string2,0        	; null terminated in-line string table
        endm   



; anywhere in you main code you can output a row  data


		nop		

        PutStr1  "Line1 Again....."         

		call	delay1sec
		call	delay1sec	
       
        PutStr2  "Line2 later....."      

		call	delay1sec
		call	delay1sec
		call	delay1sec
		call	delay1sec



;******************************************************************************

; 	Example 4   If you have fixed data to output then the above examples are ideal
;				Sometimes you may have sensors and such like bringing in new data
;               where you just want to update part of the display, though not always 
;				at that time, you might want to wait for all the sensors before
;				updating the display.
;				This routine holds 2x16 bytes of data in ram so you update the ram 
;				and then call a routine to output the ram data to the lcd.


		cblock	0x010				; 32 users register for the lcd
		T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16
		B1,B2,B3,B4,B5,B6,B7,B8,B9,B10,B11,B12,B13,B14,B15,B16
		endc


		call	row1data			; load set data into the ram memory
		call	row2data			; you can store and call as many of these as you want.

		call	row1_display		; display ram data
		call	row2_display

									; assume you are reading in some sensors or data in Ascii format
		movlw	'7'
		movwf	T14
		movlw	'8'
		movwf	T15
		movlw	'9'
		movwf	T16	
	
		call	delay1sec			; assume you carry on finishing some other functions
		call	delay1sec
		call	delay1sec

		call	row1_display		; now you want to dispay the next data, preserving the rest of the data on the row.

		call	delay1sec			
		call	delay1sec
		call	delay1sec
		call	delay1sec			
		call	delay1sec
		call	delay1sec

		goto	direct_text


;****************************************************************************
; Sub-Routines to display on a 2x16 character LCD display.
;
; LcdInit - call this subroutine before using the display. It uses
;           bits 7-1 of PORTB, leaving bit-0 free (although it will
;           be configured as an output).
;
; Lcd_C -   Sends the byte in W as a command to the display. Checks
;           if busy first. 
;
; Lcd_D -   Sends the byte in W as data. Checks for busy first.
;
;
;  Adaprted from the original code by
;  Ron Kreymborg AU  http://www.jennaron.com.au/pics/lcd.html
;
;******************************************************************




lcd_com	macro	arg1
		movlw	arg1
		call	lcd_c
		endm
	

; Flags -
lcd_Bflg	equ	0
lcd_RSflg	equ	1

; I/O portB - 
lcd_BUSY	equ	7					; input - LCD is busy (0x80)
lcd_R_W		equ	3					; output - LCD Read/Write (0x08)
lcd_RS		equ	2					; output - Register Select (0x04)
lcd_E		equ	1					; output - LCD Enable (0x02)



lcd_init
		banksel LCDPORT
		clrf	LCDPORT
		banksel	LCDTRIS
		clrf 	LCDTRIS				; all outputs and low to start
		bcf		lcd_flags,lcd_RSflg
		call	delay15				; 15mSec power up delay		
		movlw	b'00110000'			; 8-bit mode
		banksel LCDPORT
		movwf	LCDPORT
		call	lcd_clk
		call	delay5				; 4mSec wait
		movlw	b'00110000'			; 8-bit mode
		movwf	LCDPORT
		call	lcd_clk
		call	delay1				; 1 mSec wait
		movlw	b'00110000'			; 8-bit mode
		movwf	LCDPORT
		call	lcd_clk
		call	delay5					; 4mSec wait
		movlw	b'00100000'			; set for 4-bit
		movwf	LCDPORT
		call	lcd_clk

		lcd_com	b'00101000'			; 4-bits, 2-lines, 5x7
		lcd_com	b'00001000'			; display off, cursor off, blink off
		lcd_com	b'00001100'			; display on
		lcd_com	b'00000001'			; clear display
		lcd_com	b'00000110'			; increment, no display shift
		return



lcd_d	bsf		lcd_flags,lcd_RSflg	; set RS flag for data
		goto	lcd_o1


lcd_c	bcf		lcd_flags,lcd_RSflg	; clear RS flag for commands
lcd_o1	movwf	lcd_temp			; save control word
		movlw	b'11110000'			; RB7-4 as inputs for busy
		banksel	LCDTRIS
		movwf	LCDTRIS
		movlw	1 << lcd_R_W
		banksel	LCDPORT
		movwf	LCDPORT			; set up for read
lcd_o2	bcf		lcd_flags,lcd_Bflg	; assume not busy
		bsf		LCDPORT,lcd_E			; clock high
		nop							; little wait
		btfsc	LCDPORT,lcd_BUSY		; busy set?
		bsf		lcd_flags,lcd_Bflg	; yes
		bcf		LCDPORT,lcd_E			; clock low
		nop							; little wait
		nop
		call	lcd_clk				; get low bits but ignore (for now)
		btfsc	lcd_flags,lcd_Bflg	; was it busy?
		goto	lcd_o2				; yes
		banksel	LCDTRIS				; RB7-4 outputs again
		clrf	LCDTRIS
		banksel LCDPORT
		movf	lcd_temp,w			; get word to send
		call	lcd_o3				; send high nibble
		swapf	lcd_temp,w			; get word again
lcd_o3	andlw	0xf0				; just high bits (clears all control bits)
		movwf	LCDPORT
		btfsc	lcd_flags,lcd_RSflg	; was lcd_RS set?
		bsf		LCDPORT,lcd_RS		; yes
		call	lcd_clk				; write current nibble

		return

		
	
lcd_clk	bsf		LCDPORT,lcd_E			; pulse E line high
		call 	delay1
		bcf		LCDPORT,lcd_E

		return


; ***************************************************************************
;
;	Delay Sub Routines based on a 4 meg Crystal/ Oscillator
;	==================================================



delay1		movlw	0xC7				; delay 1ms  - 998 cycles
			movwf	d1
			movlw	0x01
			movwf	d2
dly1_1
			decfsz	d1, f
			goto	dly1_2
			decfsz	d2, f
dly1_2		goto	dly1_1

			return


		
delay5		movlw	0xE7			;delay 5ms  - 4998 cycles
			movwf	d1
			movlw	0x04
			movwf	d2
dly5_1
			decfsz	d1, f
			goto	dly5_2
			decfsz	d2, f
dly5_2		goto	dly5_1
			retlw	0x00

delay15		call	delay5	
			call	delay5
			call	delay5

			return



delay250	movlw	d'250'				; delay 250 ms 
			movwf	d1
dly250_1	movlw	0xC7
			movwf	d2
			movlw	0x01
			movwf	d3
dly250_2
			decfsz	d2, f
			goto	dly250_3
			decfsz	d3, f
dly250_3	goto	dly250_2
			decfsz	d1	,f
			goto	dly250_1
			retlw	0x00



delay500	call	delay250			; delay 500ms
			call	delay250
			return

delay1sec	call	delay500
			call	delay500
			return



row1data	movlw	'T'				; text data to load into ram memory row1
			movwf	T1				
			movlw	'O'
			movwf	T2
			movlw	'P'
			movwf	T3	
			movlw	' '
			movwf	T4
			movlw	'R'
			movwf	T5
			movlw	'O'
			movwf 	T6
			movlw	'W'
			movwf	T7
			movlw 	' '
			movwf	T8
			movlw 	'D'
			movwf	T9
			movlw 	'A'
			movwf	T10
			movlw	'T'
			movwf	T11
			movlw 	'A'
			movwf	T12
			movlw 	' '
			movwf	T13
			movlw	'1'
			movwf	T14
			movlw 	'2'
			movwf	T15
			movlw 	'3'
			movwf	T16
		
			return
		
		
row2data	movlw	'B'			; text data to load into ram memory row2
			movwf	B1				
			movlw	'O'
			movwf	B2
			movlw	'T'
			movwf	B3	
			movlw	'T'
			movwf	B4
			movlw	'O'
			movwf	B5
			movlw	'M'
			movwf 	B6
			movlw	' '
			movwf	B7
			movlw 	'R'
			movwf	B8
			movlw	'O'
			movwf	B9
			movlw	'W'
			movwf	B10
			movlw	' '
			movwf	B11
			movlw	'D'
			movwf	B12
			movlw	'A'
			movwf	B13
			movlw	'T'
			movwf	B14
			movlw	'A'
			movwf	B15
			movlw	' '
			movwf	B16

			return



row1_display
			movlw	0x80			; add row1 address for command line	
			call    lcd_c			; 
			movf	T1,w
			call	lcd_d
			movf	T2,w
			call	lcd_d
			movf	T3,w
			call	lcd_d
			movf	T4,w
			call	lcd_d
			movf	T5,w
			call	lcd_d
			movf	T6,w
			call	lcd_d
			movf	T7,w
			call	lcd_d
			movf	T8,w
			call	lcd_d
			movf	T9,w
			call	lcd_d
			movf	T10,w
			call	lcd_d
			movf	T11,w
			call	lcd_d
			movf	T12,w
			call	lcd_d
			movf	T13,w
			call	lcd_d
			movf	T14,w
			call	lcd_d
			movf	T15,w
			call	lcd_d
			movf	T16,w
			call	lcd_d
			
			return	

	

row2_display
			movlw	0xC0			; add Row2 address for command line	
			call    lcd_c			
			movf	B1,w
			call	lcd_d
			movf	B2,w
			call	lcd_d
			movf	B3,w
			call	lcd_d
			movf	B4,w
			call	lcd_d
			movf	B5,w
			call	lcd_d
			movf	B6,w
			call	lcd_d
			movf	B7,w
			call	lcd_d
			movf	B8,w
			call	lcd_d
			movf	B9,w
			call	lcd_d
			movf	B10,w
			call	lcd_d
			movf	B11,w
			call	lcd_d
			movf	B12,w
			call	lcd_d
			movf	B13,w
			call	lcd_d
			movf	B14,w
			call	lcd_d
			movf	B15,w
			call	lcd_d
			movf	B16,w
			call	lcd_d

			return

			end